#include <stdio.h>
#include "b_uart.h"
#include "r_cg_macrodriver.h"
#include "r_cg_serial.h"

/**
 * @brief Uart::read_completeの実装
 *
 * @param this Uart構造体へのポインタ
 */
static void Uart_read_complete(Uart *const this)
{
    this->rx_status = UART_RX_RECV_COMPLETE;
    return;
}

/**
 * @brief Uart::write_completeの実装
 *
 * @param this Uart構造体へのポインタ
 */
static void Uart_write_complete(Uart *const this)
{
    this->tx_status = UART_TX_SEND_COMPLETE;
    return;
}

/**
 * @brief Uart::error_callbackの実装
 *
 * @param this Uart構造体へのポインタ
 * @param err_type エラー種別
 */
static void Uart_error_callback(Uart *const this, uint8_t err_type)
{
    this->rx_status = UART_RX_ERROR;
    return;
}

/**
 * @brief Uart::overrun_callbackの実装
 *
 * @param this Uart構造体へのポインタ
 * @param rx_data 受信データ
 */
static void Uart_overrun_callback(Uart *const this, uint16_t rx_data)
{
    this->rx_status = UART_RX_ERROR;
    return;
}

/**
 * @brief Uart::putcharの実装
 *
 * @param this Uart構造体へのポインタ
 * @param c 書き込む文字
 * @return 成功時:書き込んだ文字  失敗時:EOF
 */
static int32_t Uart_putchar(Uart *const this, int32_t c)
{
    if (this->send(this, (uint8_t *)&c, 1U) == -1)
    {
        return (EOF);
    }
    return (c);
}

/**
 * @brief Uart::getcharの実装
 *
 * @param this Uart構造体へのポインタ
 * @return 成功時:読み込んだ文字  失敗時:EOF
 */
static int32_t Uart_getchar(Uart *const this)
{
    uint8_t data;
    if (this->recv(this, &data, 1U) == -1)
    {
        return (EOF);
    }
    return ((int32_t)data);
}

/**
 * @brief Uart::sendの実装
 *
 * @param this Uart構造体へのポインタ
 * @param tx_buf 書き込む文字へのポインタ
 * @param tx_num 書き込む文字数
 * @return 成功時:0 失敗時:-1
 */
static int32_t Uart_send(Uart *const this, uint8_t *const tx_buf, uint16_t tx_num)
{
    /* 状態を初期化 */
    this->tx_status = UART_TX_SENDING;

    /* 送信する */
    if (this->conf.func_send(tx_buf, tx_num) != MD_OK)
    {
        /* 送信エラー */
        return (-1);
    }

    /* タイムアウトを設定 */
    uint16_t timeout = this->conf.timeout;

    /* 送信完了を待つ */
    while (this->tx_status == UART_TX_SENDING)
    {
        if (timeout == 0U)
        {
            /* 送信タイムアウト(割込みが発生しなかった) */
            return (-1);
        }
        timeout--;
    }

    if (this->tx_status == UART_TX_ERROR)
    {
        /* 送信エラー */
        return (-1);
    }
    return (0);
}

/**
 * @brief Uart::recvの実装
 *
 * @param this Uart構造体へのポインタ
 * @param rx_buf 読む込む先へのポインタ
 * @param rx_num 読み込む文字数
 * @return 成功時:0 失敗時:-1
 */
static int32_t Uart_receive(Uart *const this, uint8_t *const rx_buf, uint16_t rx_num)
{
    /* 状態を初期化 */
    this->rx_status = UART_RX_RECVING;

    /* 受信する */
    if (this->conf.func_receive(rx_buf, rx_num) != MD_OK)
    {
        /* 受信エラー */
        return (-1);
    }

    /* 受信完了を待つ */
    while (this->rx_status == UART_RX_RECVING)
    {
        NOP();
    }

    if (this->rx_status == UART_RX_ERROR)
    {
        /* 受信エラー */
        return (-1);
    }
    return (0);
}

/**
 * @brief コンストラクタ
 *
 * @return Uart構造体
 */
Uart new_Uart(Uart_Conf conf)
{
    return ((Uart){
        .conf = conf,                              /* 設定 */
        .tx_status = UART_TX_INIT,                 /* UART送信状態 */
        .rx_status = UART_RX_INIT,                 /* UART受信状態 */
        .read_complete = Uart_read_complete,       /* 受信完了割込み */
        .write_complete = Uart_write_complete,     /* 送信完了割込み */
        .error_callback = Uart_error_callback,     /* エラー割込み */
        .overrun_callback = Uart_overrun_callback, /* オーバーランエラー割込み */
        .putchar = Uart_putchar,                   /* putcharの実装 */
        .getchar = Uart_getchar,                   /* getcharの実装 */
        .send = Uart_send,                         /* 送信 */
        .recv = Uart_receive,                      /* 受信 */
    });
}
