#ifndef B_UART_H
#define B_UART_H

#include <stdint.h>
#include "r_cg_macrodriver.h"

/**
 * @brief UARTの設定
 */
typedef struct
{
    uint16_t timeout; /**< 送信時の割込み発生までの待ち回数 */
    MD_STATUS (*func_send)
    (uint8_t *const tx_buf, uint16_t tx_num); /**< コード生成で用意される送信関数*/
    MD_STATUS (*func_receive)
    (uint8_t *const rx_buf, uint16_t rx_num); /**< コード生成で用意される受信関数*/
} Uart_Conf;

/**
 * @brief UART送信状態
 */
typedef enum
{
    UART_TX_INIT,          /**< 初期化済み */
    UART_TX_SENDING,       /**< 送信中 */
    UART_TX_SEND_COMPLETE, /**< 送信完了 */
    UART_TX_ERROR,         /**< 送信失敗 */
} UART_TX_Status;

/**
 * @brief UART受信状態
 */
typedef enum
{
    UART_RX_INIT,          /**< 初期化済み */
    UART_RX_RECVING,       /**< 受信中 */
    UART_RX_RECV_COMPLETE, /**< 受信完了 */
    UART_RX_ERROR,         /**< 受信失敗 */
} UART_RX_Status;

typedef struct uart Uart;

/**
 * @brief UART構造体
 */
struct uart
{
    /**
     * @brief UARTの設定
     */
    Uart_Conf conf;

    /**
     * @brief UART送信状態
     */
    volatile UART_TX_Status tx_status;

    /**
     * @brief UART受信状態
     */
    volatile UART_RX_Status rx_status;

    /**
     * @brief リード完了割込みで呼ばれるコールバック
     *
     * @param this Uart構造体へのポインタ
     */
    void (*read_complete)(Uart *const this);

    /**
     * @brief ライト完了割込みで呼ばれるコールバック
     *
     * @param this Uart構造体へのポインタ
     */
    void (*write_complete)(Uart *const this);

    /**
     * @brief エラー割込みで呼ばれるコールバック
     *
     * @param this Uart構造体へのポインタ
     * @param err_type エラー種別
     */
    void (*error_callback)(Uart *const this, uint8_t err_type);

    /**
     * @brief オーバーラン割込みで呼ばれるコールバック
     *
     * @param this Uart構造体へのポインタ
     * @param rx_data
     */
    void (*overrun_callback)(Uart *const this, uint16_t rx_data);

    /**
     * @brief 1バイトUARTに出力する
     *
     * @param this Uart構造体へのポインタ
     * @param c 書き込む文字
     * @return 成功時:書き込んだ文字  失敗時:EOF
     */
    int32_t (*putchar)(Uart *const this, int32_t c);

    /**
     * @brief 1バイトUARTにから読み込む
     *
     * @param this Uart構造体へのポインタ
     * @return 成功時:読み込んだ文字  失敗時:EOF
     */
    int32_t (*getchar)(Uart *const this);

    /**
     * @brief UARTに書き込む
     *
     * @param this Uart構造体へのポインタ
     * @param tx_buf 書き込む文字へのポインタ
     * @param tx_num 書き込む文字数
     * @return 成功時:0 失敗時:-1
     */
    int32_t (*send)(Uart *const this, uint8_t *const tx_buf, uint16_t tx_num);

    /**
     * @brief UARTから読み込む
     *
     * @param this Uart構造体へのポインタ
     * @param rx_buf 読む込む先へのポインタ
     * @param tx_num 読み込む文字数
     * @return 成功時:0 失敗時:-1
     */
    int32_t (*recv)(Uart *const this, uint8_t *const rx_buf, uint16_t rx_num);
};

/**
 * @brief コンストラクタ
 *
 * @return Uart構造体
 */
Uart new_Uart(Uart_Conf conf);

#endif
